;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-noter" "20260730.237"
  "A synchronized, Org-mode, document annotator."
  '((emacs  "24.4")
    (cl-lib "0.6")
    (org    "9.4"))
  :url "https://github.com/org-noter/org-noter"
  :commit "feae91ca4ee6fdc7fd2489fa751dd96bc1f3ddf2"
  :revdesc "feae91ca4ee6"
  :keywords '("lisp" "pdf" "interleave" "annotate" "external" "sync" "notes" "documents" "org-mode")
  :authors '(("Gonçalo Santos" . "in@bsentia")
             ("Maintainer Dmitry M" . "dmitrym@gmail.com"))
  :maintainers '(("Peter Mao" . "peter.mao@gmail.com")
                 ("Dmitry M" . "dmitrym@gmail.com")))
